from builtin import *


class Collector(CreepLogic):
    NAME = "c"
    LINK = ["wall-breaker", "scout"]
    RECIPE = [CARRY, CARRY, CARRY, MOVE, MOVE, MOVE]

    def loading(self, k: GlobalKnowledge) -> any:
        self.enable(k.sp)  # The first logic need set a spawn.

    def removed(self, k: GlobalKnowledge) -> any:
        print("remove:", self)

    def killed(self, k: GlobalKnowledge) -> any:
        self.remove()

    # --------------------------------------------------------
    def birth(self, c: Creep, k: GlobalKnowledge, *refs: CreepLogic) -> any:
        print(c, "is be created!")

    def step(self, c: Creep, k: GlobalKnowledge, *refs: CreepLogic) -> any:
        box = get.find(k.boxes, st.energetic)
        if box: res = put.carry(c, box, k.sp)


class Scout(CreepLogic):
    NAME = "scout"
    LINK = ["scout"]
    RECIPE = [MOVE, MOVE, MOVE, MOVE, MOVE, RANGED_ATTACK, MOVE, MOVE, MOVE, MOVE, MOVE, HEAL]
    OPTIMISE = False  # 禁用RECIPE自动优化
    EXTENSION = False  # 不使用extention中的能量生产该单位

    def step(self, c: Creep, k: GlobalKnowledge, *refs: CreepLogic) -> any:
        # # Select Target
        target = get.closest(c, k.battles.enemies)
        follow_range = 4

        # 如果对方没有攻击性Creep，则攻击对方基地
        if not target:
            target = k.esp
            follow_range = 3

        # 如果附近有非军事单位，且我到原目标距离较远，那么我可以先攻击它们
        close = get.closest(c, k.enemies)
        if get.distance(c, target) > 3 >= get.distance(c, close):
            put.attack(c, close, False)

        # 距离修正
        hp = get.health(c, True)  # 获取百分比生命值
        if hp > 0.9:
            follow_range -= 1
        elif hp < 0.5:
            follow_range += 1
        if st.melee(target): follow_range += 1

        # # Attack and Follow
        if target:
            put.attack(c, target, False)
            put.follow(c, target, follow_range, IGNORE_SWAMP)

        # # Heal self & near
        arounds = get.inrange(c, k.friends, 3)
        arounds = sorted(arounds, key=lambda item: get.health(item, True))
        nears = get.inrange(c, arounds, 1)  # 因为arounds已经sorted，这里不需要再sorted
        if len(nears):
            put.heal(c, nears[0], False)  # 优先治疗距离1的单位
        elif len(arounds):
            put.heal(c, arounds[0], False)  # 其次治疗范围内的单位
        else:
            put.heal(c, c, False)  # 默认治疗自己


class WallBreaker(CreepLogic):
    NAME = "wall-breaker"
    LINK = ["builder"]
    RECIPE = [ATTACK, ATTACK, ATTACK, ATTACK, MOVE, MOVE]

    def step(self, c: Creep, k: GlobalKnowledge, *refs: CreepLogic) -> any:
        wall = get.find(k.walls, st.wall)
        if wall: put.attack(c, wall)


class Builder(CreepLogic):
    NAME = "builder"
    ONCE = True  # 如果被kill了，执行killed后会自动将自己remove
    RECIPE = [WORK, WORK, CARRY, MOVE]

    def loading(self, k: GlobalKnowledge) -> any:
        self.sites = SitePlaner([  # 围绕spawn修建一圈城防
            SitePlan(k.sp, 0, 0, StructureRampart),
            SitePlan(k.sp, 0, 1, StructureRampart),
            SitePlan(k.sp, 0, -1, StructureRampart),
            SitePlan(k.sp, 1, 0, StructureRampart),
            SitePlan(k.sp, -1, 0, StructureRampart),
            SitePlan(k.sp, 1, 1, StructureRampart),
            SitePlan(k.sp, 1, -1, StructureRampart),
            SitePlan(k.sp, -1, 1, StructureRampart),
            SitePlan(k.sp, -1, -1, StructureRampart),
        ])

    def step(self, c: Creep, k: GlobalKnowledge, *refs: CreepLogic) -> any:
        site = self.sites.next()
        if site:
            put.fetch(c, k.sp)
            put.build(c, site)


# Init at tick == 1
def init(k):
    k.sp = get.spawn(st.my)
    k.esp = get.spawn(st.enemy)

    if k.sp.x < 50:
        k.side = -1
    else:
        k.side = 1

    # details
    if k.side < 0:
        k.walls = [
            get.wall(lambda w: w.x is 2 and w.y is 45),
            get.wall(lambda w: w.x is 11 and w.y is 45),
        ]
        k.boxes = [
            get.box(lambda b: b.x is 1 and b.y is 44),
            get.box(lambda b: b.x is 12 and b.y is 46),
            get.box(lambda b: b.x is 1 and b.y is 46),
            get.box(lambda b: b.x is 12 and b.y is 44),
        ]
    else:
        k.walls = [
            get.wall(lambda w: w.x is 97 and w.y is 54),
            get.wall(lambda w: w.x is 88 and w.y is 54),
        ]
        k.boxes = [
            get.box(lambda b: b.x is 98 and b.y is 55),
            get.box(lambda b: b.x is 87 and b.y is 53),
            get.box(lambda b: b.x is 98 and b.y is 53),
            get.box(lambda b: b.x is 87 and b.y is 55),
        ]

    Scheduler().add("c")  # 添加一个Collector到更新队列中


# Step at each tick >= 1
def step(k):
    print("当前的生产队列:")
    q = CreepLogic.productQueue()
    for i in q:
        print("  ", i)
    print("")
