// > sort 5
import {LANGUAGE} from "../build/main.mjs";

class _Engine_Scheduler {
    static instance = null;
    constructor() {
        if (_Engine_Scheduler.instance !== null) {
            return _Engine_Scheduler.instance;
        }
        this.prototypes = {};  // name:prototype
        this.updates = [];
        // instance
        _Engine_Scheduler.instance = this;
    }

    login(prototype) {
        if (prototype.NAME !== undefined) {
            if (this.prototypes[prototype.NAME] !== undefined) {
                if (LANGUAGE === "cn") {
                    std.warn("Scheduler.login", "名称重复:'" + prototype.NAME + "'");
                } else {
                    std.warn("Scheduler.login", "Name repeat:'" + prototype.NAME + "'");
                }
            }
            // prototype name不允许以数字结尾
            if (prototype.NAME.match(/\d+$/)) {
                if (LANGUAGE === "cn") {
                    std.error("Scheduler.login", "类型名称不允许以数字结尾:'" + prototype.NAME + "'");
                } else {
                    std.error("Scheduler.login", "Type-Name not allowed to end with a number:'" + prototype.NAME + "'");
                }
            }
        } else if (LANGUAGE === "cn") {
            std.error("Scheduler.login", "未指定名称:'NAME'");
        } else {
            std.error("Scheduler.login", "No 'NAME' specified");
        }

        this.prototypes[prototype.NAME] = prototype;

    }

    remove(name) {
        if (_Engine_Scheduler.instance === null) return;
        var _length = len(_Engine_Scheduler.instance.updates);
        for (var i = 0; i < _length; i++) {
            var item = _Engine_Scheduler.instance.updates[i];
            if (item.py_name === name) {
                item.remove();
                _Engine_Scheduler.instance.updates.splice(i, 1);
                return;
            }
        }
    }

    add(name) {
        // 根据name找到prototype，然后创建实例
        const prototype = this.prototypes[name];
        if (prototype === undefined) {
            if (LANGUAGE === "cn") {
                std.error("Scheduler.add", "找不到名称:'" + name + "'");
            } else {
                std.error("Scheduler.add", "Name not found:'" + name + "'");
            }
            return;
        }
        try{
            const instance = prototype();  //  this is a python prototype, so it can be called instead of new
            this.updates.push(instance);
            return instance;
        } catch (e) {
            if (LANGUAGE === "cn") std.warn("Scheduler.add", "创建实例失败:'" + name + "'\n", e);
            else std.warn("Scheduler.add", "Failed to create instance:'" + name + "'\n", e);
        }
    }

    find(name) {
        /// 根据名称查找实例。如果name不以数字结尾，则返回所有该类型的实例；如果name以数字结尾，则返回对应实例。
        /// .例如，'atk'返回所有'atk'类型的实例[atk0, atk1, atk2, ...]；'atk0'仅返回'atk0'实例。
        const _length = len(this.updates);
        if (name.match(/\d+$/)) {
            for (let i = 0; i < _length; i++) {
                const item = this.updates[i];
                if (item.py_name === name) {
                    return item;
                }
            }
        } else {
            const result = [];
            for (let i = 0; i < _length; i++) {
                const item = this.updates[i];
                if (item.py_name.match(new RegExp("^" + name + "\\d+$"))) {
                    result.push(item);
                }
            }
            return result;
        }

    }

    handle() {
        var _length = len(this.updates);
        if (_length > 0) {
            var removes = null;
            for (var i = 0; i < _length; i++) {
                var item = this.updates[i];
                try{
                    item.handle();
                } catch (e) {
                    if (LANGUAGE === "cn") std.warn("Scheduler", "在执行'"+ item.py_name + "'时发生错误:\n", e);
                    else std.warn("Scheduler", "An error occurred while executing '" + item.py_name + "':\n", e);
                }
                if (!item._alive){
                    item.remove();
                    if (removes === null) removes = [];
                    removes.push(i);
                }
            }
            if (removes !== null) {
                for (var i = len(removes) - 1; i >= 0; i--) {
                    this.updates.splice(removes[i], 1);
                }
            }
        }
    }
}

function Scheduler() {
    if (_Engine_Scheduler.instance !== null) return _Engine_Scheduler.instance;
    return new _Engine_Scheduler();
}

class _Engine_SpawnPlanner {
    static instance = null;
    constructor() {
        if (_Engine_SpawnPlanner.instance !== null) {
            return _Engine_SpawnPlanner.instance;
        }
        this.logics = {};
        this.tempCreeps = {};
        // instance
        _Engine_SpawnPlanner.instance = this;
    }

    require(logic) {
        if (!logic) {
            const message = LANGUAGE === "cn" ? "未提供CreepLogic对象" : "No CreepLogic object provided";
            std.error("SpawnPlanner.require", message);
            return;
        }

        if (!st.spawn(logic.spawn)) {
            const message = LANGUAGE === "cn" ? "CreepLogic对象的spawn属性不符合要求" : "The spawn property of CreepLogic object does not meet the requirements";
            std.error("SpawnPlanner.require", message);
        }

        if (logic.recipe===undefined || !logic.py_name===undefined || !logic.optimise===undefined) {
            const message = LANGUAGE === "cn" ? "CreepLogic对象缺少必要的属性recipe、name或者optimise" : "CreepLogic object is missing required properties 'recipe' 'name' or 'optimise'";
            std.error("SpawnPlanner.require", message, "result:", logic.py_name, logic.recipe, logic.optimise);
            return;
        }

        // 检查是否已经有Creep暂存
        if (this.tempCreeps[logic.py_name]) {
            const creep = this.tempCreeps[logic.py_name];
            this.tempCreeps[logic.py_name] = undefined;
            return creep;
        }

        // 调用put.create方法
        const result = put.create(logic.spawn, logic.recipe, logic.py_name, logic.optimise);
        if (result instanceof Creep) {
            // 如果结果是Creep，先检查是否是当前逻辑的Creep
            if (result.name === logic.py_name) {
                // 是当前逻辑的Creep，直接返回
                return result;
            } else {
                // 不是当前逻辑的Creep，暂存起来
                this.tempCreeps[result.name] = result;
                return ERR_BUSY;
            }
        } else if (typeof result === 'number') {
            // 如果结果是数字，表示创建失败或需要等待
            return result;
        } else {
            const message = LANGUAGE === "cn" ? "put.create返回了意外的结果" : "put.create returned an unexpected result";
            std.error("SpawnPlanner.require", message);
        }
    }
}

function SpawnPlanner() {
    if (_Engine_SpawnPlanner.instance !== null) return _Engine_SpawnPlanner.instance;
    return new _Engine_SpawnPlanner();
}
