// > sort 4
import {LANGUAGE} from "../build/main.mjs";

export var put = {
	_spawning: {},
	_spawning_real_creep: {},
	__check_unit: function (method_name, unit, method=null) {
		if (!unit) {
            if (LANGUAGE === "cn") std.warn ('put.' + method_name, '单位{unit}未定义。忽略该命令。');
			else std.warn ('put.' + method_name, 'unit is undefined. Ignore.')
            return false;
        }
        if (!unit.exists) {
			if (LANGUAGE === "cn") std.warn ('put.' + method_name, `单位{unit}:${unit.id}不存在。忽略该命令。`);
			else std.warn ('put.' + method_name, `unit:${unit.id} does not exist. Ignore.`)
            return false;
		}
		if (method && !unit[method]) {
			if (LANGUAGE === "cn") std.warn ('put.' + method_name, `单位{unit}:${unit.id}没有'${method}'方法。忽略该命令。`);
			else std.warn ('put.' + method_name, `unit:${unit.id} do not have '${method}' method. Ignore.`)
			return false;
		}
		return true;
    },
    __check_target: function (method_name, target, method=null) {
		if (!target) {
            if (LANGUAGE === "cn") std.warn ('put.' + method_name, '目标{target}未定义。忽略该命令。');
			else std.warn ('put.' + method_name, 'target is undefined. Ignore.')
            return false;
        }
        if (!target.exists) {
			if (LANGUAGE === "cn") std.warn ('put.' + method_name, `目标{target}:${target.id}不存在。忽略该命令。`);
			else std.warn ('put.' + method_name, `target:${target.id} does not exist. Ignore.`)
            return false;
		}
		if (method && !target[method]) {
			if (LANGUAGE === "cn") std.warn ('put.' + method_name, `目标{target}:${target.id}没有'${method}'方法。忽略该命令。`);
			else std.warn ('put.' + method_name, `target:${target.id} do not have '${method}' method. Ignore.`)
			return false;
		}
		return true;
    },
	move:function (unit, to, options=null) {
		if (!put.__check_unit('put.move', unit, 'moveTo')) return -1;
		if (to === null || to === undefined) {
			if (LANGUAGE === "cn") std.warn('put.move', `单位{unit}:${unit.id} -> 目标:NA。忽略。`);
			else std.warn('put.move', `unit:${unit.id} -> target:NA. Ignore.`)
			return -1;
		}
		if (to.x !== undefined && to.y !== undefined) {
			if (!(unit.path_goal) || unit.path_goal.x !== to.x && unit.path_goal.y !== to.y) {
				var path_res = searchPath (unit, to, options);
				if (path_res.incomplete) {
					if (LANGUAGE === "cn") std.warn ('put.move', `到目标的路径不完整。`);
					else std.warn ('put.move', 'Path to target:', to, ' is incomplete.');
				}
				unit.path = path_res.path;
				unit.path_goal = to;
				unit.path_index = 0;
			}
			if (unit.path) {
				to = unit.path [unit.path_index];
				unit.path_index++;
				if (unit.path_index >= unit.path.length) {
					unit.path = null;
					unit.path_index = 0;
				}
			}
			var _ = unit.moveTo (to);
		}
		else {
			var _ = unit.move (to);
		}
		if (_ === OK) {
			unit.last_move = know.now;
		}
		return _;
	},
	push:function (unit, target, options=null) {
		if (!put.__check_unit('put.push', unit, 'moveTo')) return -1;
		if (!put.__check_target('put.push', target, 'pull')) return ERR_INVALID_TARGET;
		var _ = put.move (unit, target, options);
		if (_ !== OK) {
			return _;
		}
		return target.pull (unit);
	},
	attack:function (unit, target, move=true) {
		if (!put.__check_unit('put.attack', unit, 'attack')) return -1;
		if (!put.__check_target('put.attack', target)) return ERR_INVALID_TARGET;
		if (unit instanceof StructureTower) return {attack: unit.attack(target)};
		var melee = st.melee (unit);
		var ranged = st.ranged (unit);
		var tick = know.now;
		var dist = get.distance (unit, target);
		var result = {};
		if (dist === 1) {
			if (melee) {
				result.melee = unit.attack (target);
			}
			if (ranged) {
				result.ranged = unit.rangedMassAttack ();
			}
			if (move && !melee && st.atkable (target)) {
				result.move = put.escape (unit, target, (move === true ? null : move));
			}
		}
		else if (dist <= 3) {
			if (ranged) {
				result.ranged = unit.rangedAttack (target);
				if (move && dist <= 2) {
					result.move = put.escape (unit, target, (move === true ? null : move));
				}
			}
			else if (move && melee) {
				result.move = put.move (unit, target, (move === true ? null : move));
			}
		}
		else if (move) {
			result.move = put.move (unit, target, (move === true ? null : move));
		}
		if (result.melee === OK) {
			unit.last_melee = tick;
		}
		if (result.ranged === OK) {
			unit.last_ranged = tick;
		}
		return result;
	},
	heal:function (unit, target, move=true) {
		if (!put.__check_unit('put.heal', unit, 'heal')) return -1;
		if (!put.__check_target('put.heal', target)) return ERR_INVALID_TARGET;
		if (unit instanceof StructureTower) return {heal: unit.heal(target)};
		var result = {};
		var tick = know.now;
		var dist = get.distance (unit, target);
		if (dist <= 1) {
			if (!(unit.last_melee) || unit.last_melee !== tick) {
				result.heal = unit.heal (target);
			}
			else if (!(unit.last_ranged) || unit.last_ranged !== tick) {
				result.ranged = unit.rangedHeal (target);
			}
		}
		else if (dist <= 3) {
			result.ranged = unit.rangedHeal (target);
		}
		else if (move) {
			result.move = put.move (unit, target, (move === true ? null : move));
		}
		if (result.heal === OK || result.ranged === OK) {
			unit.last_heal = tick;
		}
		return result;
	},
	fetch:function (unit, target, resource_type=null, amount=null, move=true) {
		if (!put.__check_unit('put.fetch', unit)) return -1;
		if (!put.__check_target('put.fetch', target)) return ERR_INVALID_TARGET;
		if (resource_type === null || resource_type === undefined) resource_type = RESOURCE_ENERGY;
		var tick = know.now;
		if (move && get.distance (unit, target) > 1) {
			return put.move (unit, target, (move === true ? null : move));
		}
		if((target instanceof Resource) && target.resourceType === RESOURCE_ENERGY) return unit.pickup(target);
		unit.last_fetch = tick;
		if (amount) {
			if(target instanceof Creep) return target.transfer(unit, resource_type, amount);
			if(target instanceof Structure) return unit.withdraw(target, resource_type, amount);
		}
		else {
			if(target instanceof Creep) return target.transfer(unit, resource_type);
			if(target instanceof Structure) return unit.withdraw(target, resource_type);
		}
		return ERR_INVALID_TARGET;
	},
	deposit:function (unit, target, resource_type=null, amount=null, move=true) {
		if (!put.__check_unit('put.deposit', unit)) return -1;
		// if (!put.__check_target('put.deposit', target)) return ERR_INVALID_TARGET;
		// target can be null
		if (resource_type === null || resource_type === undefined) resource_type = RESOURCE_ENERGY;
		unit.last_deposit = know.now;
		if (target === null) {
			return unit.drop (resource_type, amount);
		}
		if (move && get.distance (unit, target) > 1) {
			return put.move (unit, target, (move === true ? null : move));
		}
		if (amount) {
			if (st.storable (target)) {
				return unit.transfer (target, resource_type, amount);
			}
			return unit.drop (resource_type, amount);
		}
		else {
			if (st.storable (target)) {
				return unit.transfer (target, resource_type);
			}
			return unit.drop (resource_type);
		}
	},
	build:function (unit, site, move=true) {
		if (!put.__check_unit('put.build', unit, "build")) return -1;
		if (!put.__check_target('put.build', site)) return ERR_INVALID_TARGET;
		unit.last_build = know.now;
		if (move && get.distance (unit, site) > 3) {
			return put.move (unit, site, (move === true ? null : move));
		}
		return unit.build (site);
	},
	harvest:function (unit, target, move=true) {
		if (!put.__check_unit('put.harvest', unit, "harvest")) return -1;
		if (!put.__check_target('put.harvest', target)) return ERR_INVALID_TARGET;
		if (!st.source(target)){
			if (LANGUAGE === "cn") std.warn('put.harvest', `目标{target}:${target.id}不是资源点。忽略此命令。`);
			else std.warn('put.harvest', `target:${target.id} is not a source. Ignore.`)
			return ERR_INVALID_TARGET;
		}
		unit.last_harvest = know.now;
		if (move && get.distance (unit, target) > 1) {
			return put.move (unit, target, (move === true ? null : move));
		}
		return unit.harvest (target);
	},
	escape:function (unit, target, options=null, ratio=1.25) {
		if (!put.__check_unit('put.escape', unit)) return -1;
		if (!put.__check_target('put.escape', target)) return ERR_INVALID_TARGET;

		let friend = get.closest (get.filter(know.battles.friends, (c) => get.distance(c, unit) > 10));
		if (!friend) friend = get.spawn(st.friend);
		if (friend) return put.move (unit, friend, options);

		var direct = get.direction(target, unit);
		var directs = get.expandDirection(direct);
		for(const dir of directs){
			const pos = get.directionPoint(dir, unit);
			if (get.terrain (pos) === TERRAIN_WALL) continue;
			return put.move (unit, dir);
		}
		return put.move (unit, spawn, options);
	},
	intermit:function (unit, target, resource_type=null, options=null) {
		if (!put.__check_unit('put.intermit', unit)) return -1;
		if (resource_type === null || resource_type === undefined) resource_type = RESOURCE_ENERGY;
		var tick = know.now;
		var drops = get.resources (function (obj) {
			return obj.resourceType === resource_type && unit._last_idrop && unit._last_idrop.x === obj.x && unit._last_idrop.y === obj.y;
		});
		var free = unit.store.getFreeCapacity (resource_type);
		if (len (drops) && free > 0) {
			unit.last_intermit = know.now;
			return put.fetch (unit, drops [0], resource_type, free, false);
		}
		var current = unit.store.getUsedCapacity (resource_type);
		if (current === 0) {
			return DONE;
		}
		unit.last_intermit = tick;

		// check at here 将会允许最后一次搬运，避免empty的box导致不工作
		if (!put.__check_target('put.intermit', target)) return ERR_INVALID_TARGET;
		var dist = get.distance (unit, target);
		if (dist <= 1) {
			var _ = put.deposit (unit, target, resource_type, null, (options ? options : true));
			if (_ !== OK) {
				return _;
			}
			return DONE;
		}
		else {
			var _ = put.deposit (unit, null, resource_type, null, (options ? options : true));
			if (unit._last_idrop) {
				unit._last_idrop.x = unit.x;
				unit._last_idrop.y = unit.y;
			}
			else {
				unit._last_idrop = Point (unit.x, unit.y);
			}
			if (_ !== OK) {
				return _;
			}
			return put.move (unit, target, options);
		}
	},
	follow:function (unit, target, distance, options=null, move=true) {
		if (!put.__check_unit('put.follow', unit)) return -1;
		if (!put.__check_target('put.follow', target)) return ERR_INVALID_TARGET;
		var target_is_creep = st.creep (target);
		var __left0__ = (isinstance (distance, tuple) ? distance : tuple ([distance, 65535]));
		var stop_dist = __left0__ [0];
		var ignore_dist = __left0__ [1];
		var dist = get.distance (unit, target);
		if (target.my) {
			if (move && target_is_creep && (ignore_dist >= dist && dist >= stop_dist)) {
				var _ = put.move (target, unit, (move === true ? null : move));
				if (_ !== OK) {
					return _;
				}
			}
			return put.move (unit, target, options);
		} else {
			if (dist <= stop_dist) {
				return put.escape (unit, target, options);
			}
			return put.move (unit, target, options);
		}
	},
	carry:function (unit, src, dst, resource_type=null, options=null, intermit=false) {
		if (!put.__check_unit('put.carry', unit)) return -1;
		if (resource_type === null || resource_type === undefined) resource_type = RESOURCE_ENERGY;
		if (!(intermit)) {
			if (get.energy (unit) === 0) {
				// check at here 将会允许最后一次搬运，避免empty的box导致不工作
				if (!put.__check_target('put.carry', src)) return ERR_INVALID_TARGET;
				if (get.distance (unit, src) <= 1) {
					put.fetch (unit, src, resource_type, null, false);
					return put.move (unit, dst, options);
				}
				return put.fetch (unit, src, resource_type, null, (options ? options : true));
			}
			if (get.distance (unit, dst) <= 1) {
				put.deposit (unit, dst, resource_type, null, false);

				// check at here 将会允许最后一次搬运，避免empty的box导致不工作
				if (!put.__check_target('put.carry', src)) return ERR_INVALID_TARGET;
				return put.move (unit, src, options);
			}
			return put.deposit (unit, dst, resource_type, null, (options ? options : true));
		}
		else {
			var drops = get.resources (function (obj) {
				return obj.resourceType === resource_type && unit._last_idrop && unit._last_idrop.x === obj.x && unit._last_idrop.y === obj.y;
			});
			var free = unit.store.getFreeCapacity (resource_type);
			if (len (drops) && free > 0) {
				return put.fetch (unit, drops [0], resource_type, free, false);
			}

			var current = unit.store.getUsedCapacity (resource_type);
			if (current === 0) {
				// check at here 将会允许最后一次搬运，避免empty的box导致不工作
				if (!put.__check_target('put.carry', src)) return ERR_INVALID_TARGET;
				return put.fetch (unit, src, resource_type, null, (options ? options : true));
			}
			unit.last_intermit = know.now;

			var dist = get.distance (unit, dst);
			if (dist <= 1) {
				var _ = put.deposit (unit, dst, resource_type, null, (options ? options : true));
				if (_ !== OK) {
					return _;
				}
				// check at here 将会允许最后一次搬运，避免empty的box导致不工作
				if (put.__check_target('put.carry', src)) put.move (unit, src, options);
				return DONE;
			}
			else {
				var _ = put.deposit (unit, null, resource_type, null, (options ? options : true));
				if (unit._last_idrop) {
					unit._last_idrop.x = unit.x;
					unit._last_idrop.y = unit.y;
				}
				else {
					unit._last_idrop = Point (unit.x, unit.y);
				}
				var _ = put.deposit (unit, null, resource_type, null, (options ? options : true));
				if (_ !== OK) {
					return _;
				}
				return put.move (unit, dst, options);
			}
		}
	},
	site:function (x, y, building_type, single=false) {
		if (single) {
			var exists = get.site (function (ste) {
				return ste.x === x && ste.y === y && ste.structureType === building_type;
			});
			if (exists) {
				return ERR_NAME_EXISTS;
			}
		}
		var _ = createConstructionSite (x, y, building_type);
		if (_.object !== undefined) {
			return _.object;
		}
		return _.error;
	},
	create:function (spawn, recipe, creep_name=null, optimise=true) {

		// Get Record
		const spawning = put._spawning[spawn.id];
		if (spawn.spawning && spawn.spawning.remainingTime >= 2) {
			if (spawning === undefined) return ERR_BUSY;  // busy and not me.
			else if (spawning === null) {
				put._spawning[spawn.id] = spawn.spawning;
			} else {
				if (spawning.creep.id !== spawn.spawning.creep.id) return ERR_BUSY;  // busy and may not this spawn.
				put._spawning[spawn.id] = spawn.spawning;  // update
			}
			return spawn.spawning.remainingTime - 1;
		}else{
			var creep = null;
			if (spawning !== undefined && spawning !== null) {
				creep = put._spawning_real_creep[spawn.id];
				delete put._spawning_real_creep[spawn.id];
				delete put._spawning[spawn.id];
				creep.__spawn__ = spawn;
				check_and_set_stringfn(creep);
				return creep;
			} else {
				// check busy
				if (spawn.next_spawn >= know.now)
					return ERR_BUSY;
				spawn.next_spawn = know.now + CREEP_SPAWN_TIME - 1;

				// check recipe
				if (optimise) recipe = get.partsOptimise(recipe);
				var result = spawn.spawnCreep (recipe);
				if (result.error !== undefined) return result.error;
				if (creep_name) result.object.name = creep_name;
				put._spawning_real_creep[spawn.id] = result.object;
				put._spawning[spawn.id] = null;  // will be set to the spawning at end of this tick .
				return len(recipe) * CREEP_SPAWN_TIME - 1;  // ok but not a creep.
			}
		}
	},
	broadcast: function (arr, func=(x, y, val)=>val, start={x:0, y:0}, end=undefined) {
		/// 允许对一个2d数组中指定区域的元素进行逐个操作，并将返回值存储在原数组中。
		if (end === undefined) end = get.shape(arr);
		for (let x = start.x; x < end.x; x++) {
			for (let y = start.y; y < end.y; y++) {
				arr[y][x] = func(x, y, arr[y][x]);
			}
		}
	}
}
