/**
 * @file st.js
 * @description 提供各种类型判断和对象能力判断的工具函数
 * @author EagleBaby
 * @version 1.0
 * @date 2025-01-10
 */
// > sort 1
const st = {
    jnumber:    obj => typeof obj === 'number',
    jlist:      obj => obj instanceof Array,
    jstring:    obj => typeof obj === 'string',
    jbool:      obj => typeof obj === 'boolean',
    jdict:      obj => obj instanceof Object,
    na:         obj => obj === null || obj === undefined,
    // -------------------------------------- PY基础类型判断 | PY Basic Type Judgment -----------------
    list:       obj => isinstance(obj, list),
    tuple:      obj => isinstance(obj, tuple),
    dict:       obj => isinstance(obj, dict),
    set:        obj => isinstance(obj, set),
    str:        obj => isinstance(obj, str),
    int:        obj => isinstance(obj, int),
    float:      obj => isinstance(obj, float),
    bool:       obj => isinstance(obj, bool),
    // -------------------------------------- 游戏对象类型判断 | Game Object Type Judgment -----------------
    point:     obj => obj?.x !== undefined && obj?.y !== undefined,
    site:      (obj, check_exists=true) => !(check_exists && !obj?.exists) && obj instanceof ConstructionSite,
    creep:     (obj, check_exists=true) => !(check_exists && !obj?.exists) && obj instanceof Creep,
    gobject:   (obj, check_exists=true) => !(check_exists && !obj?.exists) && obj instanceof GameObject,
    ostructure:(obj, check_exists=true) => !(check_exists && !obj?.exists) && obj instanceof OwnedStructure,
    resource:  (obj, check_exists=true) => !(check_exists && !obj?.exists) && obj instanceof Resource,
    source:    (obj, check_exists=true) => !(check_exists && !obj?.exists) && obj instanceof Source,
    structure: (obj, check_exists=true) => !(check_exists && !obj?.exists) && obj instanceof Structure,
    box:       (obj, check_exists=true) => !(check_exists && !obj?.exists) && obj instanceof StructureContainer,
    extension: (obj, check_exists=true) => !(check_exists && !obj?.exists) && obj instanceof StructureExtension,
    rampart:   (obj, check_exists=true) => !(check_exists && !obj?.exists) && obj instanceof StructureRampart,
    road:      (obj, check_exists=true) => !(check_exists && !obj?.exists) && obj instanceof StructureRoad,
    spawn:     (obj, check_exists=true) => !(check_exists && !obj?.exists) && obj instanceof StructureSpawn,
    tower:     (obj, check_exists=true) => !(check_exists && !obj?.exists) && obj instanceof StructureTower,
    wall:      (obj, check_exists=true) => !(check_exists && !obj?.exists) && obj instanceof StructureWall,
    flag:      (obj, check_exists=true) => !(check_exists && !obj?.exists) && obj instanceof Flag,
    // -------------------------------------- 对象能力判断 | Object Ability Type Judgment -----------------
    spawning:  (obj, check_exists=true) => !(check_exists && !obj?.exists) && obj?.hits === undefined && obj instanceof Creep,
    movable:   (obj, check_exists=true) => !(check_exists && !obj?.exists) && obj?.body?.some(p => p.type === MOVE),
    atkable:   (obj, check_exists=true) => !(check_exists && !obj?.exists) && ((obj?.body?.some(p => p.type === ATTACK || p.type === RANGED_ATTACK)) || obj instanceof StructureTower),
    melee:     (obj, check_exists=true) => !(check_exists && !obj?.exists) && obj?.body?.some(p => p.type === ATTACK),
    ranged:    (obj, check_exists=true) => !(check_exists && !obj?.exists) && ((obj?.body?.some(p => p.type === RANGED_ATTACK)) || obj instanceof StructureTower),
    healable:  (obj, check_exists=true) => !(check_exists && !obj?.exists) && ((obj?.body?.some(p => p.type === HEAL)) || obj instanceof StructureTower),
    hitable:   (obj, check_exists=true) => !(check_exists && !obj?.exists) && obj?.hits !== undefined,
    workable:  (obj, check_exists=true) => !(check_exists && !obj?.exists) && obj?.body?.some(p => p.type === WORK),
    storable:  (obj, check_exists=true) => !(check_exists && !obj?.exists) && ((obj instanceof Structure && obj.store) || (obj instanceof Creep && obj?.body?.some(p => p.type === CARRY))),
    energetic: (obj, check_exists=true) => !(check_exists && !obj?.exists) && ((obj instanceof Structure && obj.store && get.energy(obj) > 0) || (obj instanceof Creep && get.energy(obj) > 0) || (obj instanceof Resource && obj.amount > 0)),
    damaged:   (obj, check_exists=true) => !(check_exists && !obj?.exists) && obj.hits < obj.hitsMax,
    empty:     (obj, check_exists=true) => !(check_exists && !obj?.exists) && (!obj || !obj.store || get.energy(obj) === 0),
    full:      (obj, check_exists=true) => !(check_exists && !obj?.exists) && (!obj || !obj.store || get.energy(obj, true)),
    friend:    (obj, check_exists=true) => obj && obj.my && (!check_exists || obj?.exists),
    my:        (obj, check_exists=true) => obj && obj.my && (!check_exists || obj?.exists),
    enemy:     (obj, check_exists=true) => obj && !obj.my && (!check_exists || obj?.exists),
};
