
class __View {
    static DEFAULT_OPTS = {
        font: '0.5',
        align: 'center',
        opacity: 0.7,
        color: '#AEFC80'
    }
    constructor(layer, persistent = true) {
        this.v = new Visual(layer, persistent);
    }

    clear() {
        this.v.clear(); // > ignore
        return this;
    }

    text(text, pos, font, color, options = null) {
        options = options? options : {};
        options.font = font;
        options.color = color;
        this.v.text(text, pos, options);
        return this;
    }

    circle(pos, radius, fill = null, options = null) {
        options = options? options : {};
        options.radius = radius;
        options.fill = fill;
        this.v.circle(pos, options);
        return this;
    }

    line(pos1, pos2, width = 0.1, color = null, options = null) {
        options = options? options : {};
        options.width = width;
        options.color = color;
        this.v.line(pos1, pos2, options);
        return this;
    }

    rect(pos, width, height, fill = null, options = null) {
        options = options? options : {};
        options.fill = fill;
        this.v.rect(pos, width, height, options);
        return this;
    }

    poly(points, fill = null, options = null) {
        options = options? options : {};
        options.fill = fill;

        let array = [];
        for (let i = 0; i < points.length; i++) {
            array.push(points[i]);
        }

        this.v.poly(array, options);
        return this;
    }

    size() {
        return this.v.size();
    }

    header(creep) {
        if (creep) {
            let txt = creep.name || creep.id;
            let hp_percent = get.health(creep, true);  // int:0 - 100
            let _hex = Math.floor(hp_percent * 255 / 100).toString(16).padStart(2, '0');
            let color = `#AE${_hex}80`;
            txt += `|${hp_percent}`;
            this.text(txt, get.offset(creep, 0, -0.6), 0.4, color, __View.DEFAULT_OPTS);
        } else {
            std.warn("View.header", "No creep object:", creep);
        }
        return this;
    }
}

function View(layer, persistent = true) {
    return new __View(layer, persistent);
}
